/*===============================================================================*/
/* Determining TAR order.                                                        */
/* Test used - Hansen's bootstrap                                                */
/* minimal regime size requirement 10%                                           */
/*===============================================================================*/

declare _trim, _boot, _p, _delay,_mrs;
cls;

chdir c:\TARSpec;
load qt=citibase;

y = trimr(qt,1,0) - trimr(lag1(qt),1,0);

 output file=citibase-hans-longdiff.out reset;

output on;
screen on;

algus=time;
format /ro 8,4 ; 
 outwidth 80;
  
i=1;
repl=1;

T_set=rows(y);                           /* set of sample sizes */ 

                
_trim=0.10;
_boot=199;
_delay=1;                              /* lag determining the threshold variable */
_p=12;                                  /* AR order */


   breaks=miss(zeros(repl,3),0);
  
   p_f1234=miss(zeros(repl,3),0);  /* (repl x 3) because only one T and one level */

 
  print "**************************************************** ";
  print "                               T=" (T_set-_p);
  print"                            _boot=" _boot;
  print "**************************************************** "; 
  "REPL       p_F12   p_F23   p_F34  mhat10    mhat05    mhat01  valid  total ";
  "-------   ------   ------  -----  ------    ------    ------  -----  ----- "; 

 
        T=rows(y);
        ylag=zeros(T-_p,_p);
                
        i=1;
       do while i<=_p;
          ylag[.,i]=y[_p-i+1:t-i,1];                  /* All lags */
          i=i+1;
       endo;

       _y_=y[_p+1:t,1];
       _mrs=round(_trim*rows(_y_));  /* min # of obs in a regime */
       
       yks=ones(rows(_y_),1);
        _X_=yks~ylag;
          
           
         tvarset=(cumsumc(_X_[.,2:cols(_X_)]'))'; 
        
                
    do  while _delay<=_p;
     "DELAY=" (_delay+1);
     "";   
        _tvar_=tvarset[.,_delay];                       /* transition variable */
        
       /* Fitting Linear Model (Both GP and Hansen, not using White s.e.) */
        
        mi=invpd(_X_'_X_);
        beta=mi*(_X_'_y_);  /*"Linear beta" beta';*/
        e=_y_-_X_*beta;
        ee=e'e;
        sigsq=e'e/rows(_y_);

       /* Fitting Threshold Model */
        q=_tvar_;                   /* FIX thresh.var. */
        qq=qsort(q);                  /* search range */

        gn=rows(qq);                       
        qn=cols(q);

        s=zeros(gn,qn);
        qm=q; 
        j=1;
        do while j<=gn;
          gg=qq[j];
          if scalmiss(gg); 
             continue; 
          endif;
          xd=_X_.*(qm.<=gg);
          xxd=moment(xd,0);
          xde=xd'e;
          mmi=invpd(xxd-xxd*mi*xxd);
          s[j]=xde'mmi*xde;
         j=j+1;
        endo;
                
        f=maxc(s);
        
        f=T*(f./(ee-f)); /* ee comes from linear */
        ghat=qq[maxindc(s)];
        
       
         /*"Lin.res range=" minc(e)~maxc(e);*/  /*1.level*/ 
       p_f1234[1,1]=F12_TEST(_y_,beta,e,f);
                                      /* If linearity IS rejected */
           breaks[1,1]=ghat;          /* If valid break, take into account */
          
           @ SETAR(2) Model @
           d1=(_tvar_.<=ghat);
           d2=1-d1; 
           xx=(_X_.*d1)~(_X_.*d2);
           mxx=invpd(moment(xx,0));
           beta_tar1=mxx*(xx'_y_);
           b1=beta_tar1[1:_p+1];
           b2=beta_tar1[_p+2:2*_p+2];
    
           
           "ghat1=" ghat;
           format /m1 /re 5,3;
           "F12 (pval)" f~p_f1234[1,1];
           format /ro 18,10;
           "F12 (pval)" f~p_f1234[1,1];
           
           
           
           if  p_f1234[1,1]>=0.05;
           goto loopend;
           endif;
                             
         /* Iterate ghat1 and ghat2 */
           etar1=_y_-xx*beta_tar1;
           ee_tar=etar1'etar1;
           "SSE_TAR=" ee_tar; 
                            
           ghat2=tar_d(_y_,_X_,x1,_tvar_,qsort_2(_tvar_,ghat));  @ fit TAR-2 model @
                    
           x1=_X_.*(_tvar_.<=ghat2);
           ghat1=tar_d(_y_,_X_,x1,_tvar_,qsort_2(_tvar_,ghat2));  @ iterated estimate of gamma_1 @
                      
           x1=_X_.*(_tvar_.<=ghat1);
           ghat2=tar_d(_y_,_X_,x1,_tvar_,qsort_2(_tvar_,ghat1));  @ iterated estimate of gamma_2 @
         
           gg=sortc(ghat1|ghat2,1);
           ghat1=gg[1];
           ghat2=gg[2];  
            
           "Iterated g1 g2 " ghat1~ghat2; 
           /*"distance=" abs(indexcat(sortc(_tvar_,1),ghat1)-indexcat(sortc(_tvar_,1),ghat2));
           */                    
          
          d1=(_tvar_.<=ghat1);
          d2=(_tvar_ .>ghat1 .and _tvar_.<=ghat2 );
          d3=(1-d1-d2); 
          xx=(_X_.*d1)~(_X_.*d2)~(_X_.*d3);
         
         /* "_tvar_[1:10]~Sample split [1:10]" _tvar_[1:10]~xx[1:10,.];*/
         
          /* Final SETAR(3) beta-estimates*/
         gehat=sortc(ghat1|ghat2,1); /* vector of current breaks */
         beta_tar2=invpd(moment(xx,0))*(xx'_y_);
         b21=beta_tar2[1:_p+1];
         b22=beta_tar2[_p+2:2*_p+2];     
         b23=beta_tar2[2*_p+3:rows(beta_tar2)];
         /* "beta21=" b21';"beta22=" b22'; "beta23=" b23';*/
          
         etar2=_y_-xx*(_y_/xx); 
         f23=T*((etar1'etar1)/(etar2'etar2)-1);
         
        /*"TAR(2).res range=" minc(etar1)~maxc(etar1);
        "TAR(3).res range=" minc(etar2)~maxc(etar2);*/
         
      /*2.level*/
      ftout=F23_TEST(_y_,b1,b2,ghat,etar1,f23);
      p_f1234[1,2]=ftout[1];
      rat=ftout[2:3];
      
      format /m1 /re 5,3;
      " F23 (p_F23)" f23~p_f1234[1,2];
      format /ro 18,10;
           "F23 (pval)" f~p_f1234[1,2];
      
      
      format /ro 8,4;
      if  p_f1234[1,2]>=0.05/2;
           goto loopend;
      endif; 
       
       
           
      /* Ghat3 iteration */
  
       ghat3=tar_d3(_y_,_X_,gehat,_tvar_,qsort_3(_tvar_,gehat));  @ fit TAR-3 model @
       x1=_X_.*(_tvar_.<=ghat3);
       ghat31=tar_d(_y_,_X_,x1,_tvar_,qsort_2(_tvar_,ghat3)); @ iterate gamma1 @                       
       g2hat=sortc(ghat31|ghat3,1);
       ghat32=tar_d3(_y_,_X_,g2hat,_tvar_,qsort_3(_tvar_,g2hat));  @ iterate gamma_2 @
       g2hat=sortc(ghat31|ghat32,1);
       ghat33=tar_d3(_y_,_X_,g2hat,_tvar_,qsort_3(_tvar_,g2hat)); @ iterate gamma3 @ 
       geehat=sortc(ghat31|ghat32|ghat33,1);
       ghat1=geehat[1]; 
       ghat2=geehat[2]; 
       ghat3=geehat[3];
            
       "Iterated g1 g2 g3 " ghat1~ghat2~ghat3; 
      
         
       d1=(_tvar_.<=ghat1);
       d2=(_tvar_.>ghat1 .and _tvar_.<=ghat2);
       d3=(_tvar_.>ghat2 .and _tvar_.<=ghat3);
       d4=(1-d1-d2-d3); 
       xx=(_X_.*d1)~(_X_.*d2)~(_X_.*d3)~(_X_.*d4);
       etar3=_y_-xx*(_y_/xx); 
       f34=T*((etar2'etar2)/(etar3'etar3)-1);
       
       
       /*"_tvar_[1:10]~Sample split [1:10]" _tvar_[1:10]~xx[1:10,.];*/
       /*"betas" reshape(_y_/xx,4,11);*/
       
     /*2.level*/
                      
      p_f1234[1,3]=F34_TEST(_y_,b21,b22,b23,gehat,etar2,f34);
      
      ""; 
      format /m1 /re 5,3;
      "F34 (p_F34)" f34~p_f1234[1,3];
      format /ro 8,4 ;         
loopend:       
_delay=_delay +1;
endo; /* end cycle over threshold lags*/ 
   
                  


ttime=time_com(algus);

"
Time wasted on this simulation=" ttime;
output off;
end;  
                    
           
/**************************************************************************/
/***************************** PROCEDURES *********************************/
/**************************************************************************/ 
/* =============== PROC F34_TEST  ================== */
/*     Model-Based Homoskedastic Error Bootstrap     */
/* INPUT - y,X, trans.variable                       */
/* OUTPUT - decision after bootstrapping Hansens F12 */
/*           and corresponding bootstrap p-value     */
/* ================================================= */
proc(1)=F34_TEST(yy,beta1,beta2,beta3,ghat,e3,ef3);
local y0, btpv34,fb2,i,simout,cheat,name;
 
y0=y[1:_p]; 
cheat=0;
    
  fb2=miss(zeros(_boot,1),0);
   i=_boot; 
  
  do while i ge 1;
    /*"boot=" i;*/
     simout=tar2_sim(y0,beta1,beta2,beta3,ghat,e3);
     
     if simout[1]==1;
         cheat=cheat+1;
        goto bootrep;
     else;
         fb2[i]=tar3(simout[2:rows(simout)]);
    endif;
   
    i=i-1;
   bootrep: 
   
  if cheat> 5000;
   "More than 5000 extra bootstraps for F34" ; 
     break;                    /* TOO MANY not valid bootstraps => JUMP OUT */
   endif;
 endo;
 
 /*fb2=missrv(fb2 , meanc(packr(fb2)) );*/   /* impute missing values with mean */
 
 /*"F34_b range" minc(fb2)~maxc(fb2);  */  
 btpv34=meanc(fb2.>ef3);

retp(btpv34);
endp;



/* =============== PROC F23_TEST  ================== */
/*     Model-Based Homoskedastic Error Bootstrap     */
/* INPUT - y,X, trans.variable                       */
/* OUTPUT - decision after bootstrapping Hansens F12 */
/*           and corresponding bootstrap p-value     */
/* ================================================= */
proc(1)=F23_TEST(yy,beta1,beta2,ghat,e,ef);
local y0, btpv23,fb1,i,simout,cheat,valid,rat;
 
y0=y[1:_p]; 
cheat=0;
valid=0;
    
  fb1=miss(zeros(_boot,1),0);
   i=_boot; 

  do while i ge 1;
    /*"boot=" i;*/
     simout=tar_sim1(y0,beta1,beta2,ghat,e);
     
     if simout[1]==1;
         cheat=cheat+1;
         goto bootrep;
     else;
         valid=valid+1; 
         fb1[i]=tar2(simout[2:rows(simout)]);
    endif;
    
    i=i-1;
   bootrep: 
   
  /* if cheat> 5000;
    "More than 5000 extra bootstraps for F23" ;
     break;                    /* TOO MANY not valid bootstraps => JUMP OUT */
   endif;*/
  endo;
  
   /*"f23b" fb1; 
   fb1=missrv(fb1 , meanc(packr(fb1)) ); */  /* impute missing values with mean */
 
 /* "F23_b range" minc(fb1)~maxc(fb1); */
 btpv23=meanc(fb1.>ef);

retp(btpv23|valid|cheat);
endp;

/* =============== PROC F12_TEST  ================== */
/*     Model-Based Homoskedastic Error Bootstrap     */
/* INPUT - y,X, trans.variable                       */
/* OUTPUT - decision after bootstrapping Hansens F12 */
/*           and corresponding bootstrap p-value     */
/* ================================================= */
proc(1)=F12_TEST(yy,beta,e,ef);
local y0, d_f12,btpv12,fb,i;
 
 y0=y[1:_p];     
  fb=miss(zeros(_boot,1),0);
  i=1; 
  do while i<=_boot;
    fb[i]=tar(ar_sim(y0,beta,e));
    i=i+1;
  endo;
  
 /*"F12_b range" minc(fb)~maxc(fb);*/
 btpv12=meanc(fb.>ef);
  
retp(btpv12);
endp;

/*------------------------ ERINEVAD TAR PROTSEDUURID--------------*/
proc tar3(dat);
local y,x,j,q,ghat,ghat1,ghat2,ghat3,gehat,ghat31,ghat32,ghat33,
xx,e,qd,x1,f,gg,etar4,n,d1,d2,d3,d4, etar3,dbl;
  
/* "------------------ TAR 3 sees -------------------";*/
  n=rows(dat); 
  y=dat[_p+1:n]; 
  x=ones(rows(y),1)~dat[_p:n-1];
  j=2; 
  do while j<=_p;
    x=x~dat[_p+1-j:n-j];
  j=j+1;
  endo;
   
 dbl=(cumsumc(x[.,2:cols(x)]'))'; /* y_{t-1} - y_{t-d} */
  
  qd =dbl[.,_delay];  
  ghat=tarm(y,x,qd,qsort(qd));   @ fit TAR-1 model @
  /*"tar(2) ghat" ghat;   */
  x1=x.*(qd.<=ghat);  
  xx=x1~x;
  e=y-xx*(y/xx);    
  /*"betas" reshape(y/xx,2,11);*/
  
  ghat2=tar_d(y,x,x1,qd,qsort_2(qd,ghat));  @ fit TAR-2 model @
  x1=x.*(qd.<=ghat2);
  ghat1=tar_d(y,x,x1,qd,qsort_2(qd,ghat2));  @ iterated estimate of gamma_1 @
  x1=x.*(qd.<=ghat1);
  ghat2=tar_d(y,x,x1,qd,qsort_2(qd,ghat1));  @ iterated estimate of gamma_2 @
  
  
  gehat=sortc(ghat1|ghat2,1); 
  d1=(qd.<=gehat[1]);
  d2=(qd.>gehat[1] .and qd.<=gehat[2]);
  d3=(1-d1-d2);
  xx=(x.*d1)~(x.*d2)~(x.*d3); 
  /*xx; */
  etar3=y-xx*(y/xx);    
 /* "betas" reshape(y/xx,3,11);*/
  
   /* Ghat3 iteration */
           
  ghat3=tar_d3(y,x,gehat,qd,qsort_3(qd,gehat));  @ fit TAR-3 model @
  x1=x.*(qd.<=ghat3);
  ghat31=tar_d(y,x,x1,qd,qsort_2(qd,ghat3)); @ iterate gamma1 @                       
  g2hat=sortc(ghat31|ghat3,1);
  ghat32=tar_d3(y,x,g2hat,qd,qsort_3(qd,g2hat));  @ iterate gamma_2 @
  g2hat=sortc(ghat31|ghat32,1);
  ghat33=tar_d3(y,x,g2hat,qd,qsort_3(qd,g2hat)); @ iterate gamma3 @ 
  geehat=sortc(ghat31|ghat32|ghat33,1);
  ghat1=geehat[1]; ghat2=geehat[2]; ghat3=geehat[3];
            
  d1=(qd.<=ghat1);
  d2=(qd.>ghat1 .and qd.<=ghat2);
  d3=(qd.>ghat2 .and qd.<=ghat3);
  d4=(1-d1-d2-d3); 
  xx=(x.*d1)~(x.*d2)~(x.*d3)~(x.*d4);
  etar4=y-xx*(y/xx);
  
  f=t*((etar3'etar3)/(etar4'etar4)-1);
  /*"f34-boot=" f;*/
retp(f);
endp;




proc tar2(dat);
local y,x,j,q,ghat,ghat1,ghat2,xx,e,qd,x1,f,gg,etar,n,dbl;
  
 /* " ====== simulate TAR(2) ========";*/
  n=rows(dat); 
  y=dat[_p+1:n]; 
  x=ones(rows(y),1)~dat[_p:n-1];
  j=2; 
  do while j<=_p;
    x=x~dat[_p+1-j:n-j];
  j=j+1;
  endo;
   
  dbl=(cumsumc(x[.,2:cols(x)]'))'; /* y_{t-1} - y_{t-d} */
  
  qd =dbl[.,_delay];  
   ghat=tarm(y,x,qd,qsort(qd));   @ fit TAR-1 model @
   /*" TAR2ghat =" ghat; */
     
  x1=x.*(qd.<=ghat);  /*"left size" sumc(qd.<=ghat);*/
  xx=x1~x; /*qd~xx;*/
  e=y-xx*(y/xx);    
  /*"TAR(2).res range="minc(e)~maxc(e); */ 
  
  ghat2=tar_d(y,x,x1,qd,qsort_2(qd,ghat));  @ fit TAR-2 model @
  x1=x.*(qd.<=ghat2);
  ghat1=tar_d(y,x,x1,qd,qsort_2(qd,ghat2));  @ iterated estimate of gamma_1 @
  x1=x.*(qd.<=ghat1);
  ghat2=tar_d(y,x,x1,qd,qsort_2(qd,ghat1));  @ iterated estimate of gamma_2 @
  if ghat2<ghat1; gg=ghat1;ghat1=ghat2;ghat2=gg;endif;
  
  /*"F23 sim ghat1~ghat2" ghat1~ghat2;
  "distance=" abs(indexcat(sortc(qd,1),ghat1)-indexcat(sortc(qd,1),ghat2)); */  
  d1=(qd.<=ghat1);
  d2=(qd.<=ghat2).*(1-d1);
  xx=(x.*d1)~(x.*d2)~(x.*(1-d1-d2));
  etar=y-xx*(y/xx);
  
  /*"F23 sim, tar(3) residual range" minc(etar)~maxc(etar);
  "betas" reshape((y/xx),3,3);*/
  
  
  f=t*((e'e)/(etar'etar)-1);
  
/* "f23-boot=" f;*/
retp(f);
endp;



proc tar(dat);
local y,x,q,qq,mi,e,qn,gn,t,s,n,m,qm,j,xd,xde,f,xxd,gg,mmi,dbl;
  
  n=rows(dat);
  y=dat[_p+1:n]; 
  t=rows(y);
  x=ones(t,1)~dat[_p:n-1];
  
  j=2; 
  do while j<=_p;
    x=x~dat[_p+1-j:n-j];
  j=j+1;
  endo;
  
  dbl=(cumsumc(x[.,2:cols(x)]'))'; /* y_{t-1} - y_{t-d} */
  
  q =dbl[.,_delay]; 
  
  qq=qsort(q);
  
  mi=invpd(moment(x,0));
  e=y-x*mi*(x'y);
  t=rows(e);
  qn=cols(q);
  gn=rows(qq);
  s=zeros(gn,qn);
  m=1;
  do while m<=qn;
   qm=q[.,m]; 
    j=1;do while j<=gn;
      gg=qq[j,m];
      if scalmiss(gg); continue; endif;
      xd=x.*(qm.<=gg);
      xxd=moment(xd,0);
      mmi=xxd-xxd*mi*xxd;
      xde=xd'e;
      s[j,m]=xde'dinv(xde,mmi);
    j=j+1;endo;
  m=m+1;endo;
  f=maxc(s);
  f=t*(f./((e'e)-f));
retp(f);
endp;

proc tarm(y,x,q,qq);
local mi,e,qn,gn,s,m,qm,j,xd,xde,xxd,gg,mmi,dhat,ghat;
  mi=ginv(moment(x,0));  
  e=y-x*mi*(x'y);
  qn=cols(q); 
  gn=rows(qq); 
  s=zeros(gn,qn);
  m=1; do while m<=qn;
   qm=q[.,m]; 
    j=1;do while j<=gn;
      gg=qq[j,m];
      if scalmiss(gg); continue; endif;
      xd=x.*(qm.<=gg);
      xxd=moment(xd,0);
      mmi=xxd-xxd*mi*xxd; 
      xde=xd'e;
      s[j,m]=xde'dinv(xde,mmi); 
    j=j+1;endo;
  m=m+1;endo;
  ghat=qq[maxindc(s)]; 
retp(ghat);
endp;

/*==================================*/
/*          TAR-D                   */
/*==================================*/
proc tar_d(y,x,x1,q,qq);
local mi,e,qn,gn,s,j,xd,xde,xxd,gg,mmi,ghat,xdx;
  
  mi=ginv(moment(x,0));
  e=y-x*mi*(x'y);
    
  gn=rows(qq);
  s=zeros(gn,1);
    j=1;do while j<=gn;
      gg=qq[j];
      if scalmiss(gg); continue; endif;
      xd=x1~(x.*(q.<=gg)); 
      xxd=moment(xd,0);
      xdx=x'xd;
      mmi=xxd-xdx'mi*xdx;
      xde=xd'e;
      s[j]=xde'dinv(xde,mmi);
    j=j+1;endo;
  ghat=qq[maxindc(s)];
retp(ghat);
endp;

proc tar_d3(y,x,ghat,q,qq);
local mi,e,qn,gn,s,j,xd,xde,xxd,gg,mmi,ghat3,xdx,x1,x2,x3;
  
  mi=ginv(moment(x,0));
  e=y-x*mi*(x'y);
   
  gn=rows(qq);
  s=zeros(gn,1);
  j=1;
  do while j<=gn;
    gg=qq[j];
    if scalmiss(gg); continue; endif;
    ghat3=sortc(ghat|gg,1); /*"ghat3=" ghat3;*/
      
    x1=x.*(q.<=ghat3[1]);
    x2=x.*(q.>ghat3[1] .and q.<=ghat3[2]);
    x3=x.*(q.>ghat3[2] .and q.<=ghat3[3]); 
      
    xd=x1~x2~x3; 
    xxd=moment(xd,0);
    xdx=x'xd;
    mmi=xxd-xdx'mi*xdx;
    xde=xd'e;
    s[j]=xde'dinv(xde,mmi);
    j=j+1;
  endo;
  ghat3=qq[maxindc(s)];
retp(ghat3);
endp;



@********************** Simulation Procedures ******************* @

proc ar_sim(y0,beta,e);
local t,u,yb,k;
  
  t=rows(e);
  k=rows(beta);
  u=e[ceil(rndu(t,1)*t)]+beta[1];
  yb=recserar(y0|u,y0,beta[2:k]);
 
retp(yb); 
endp;

/* =============== TAR_SIM1 ===============*/
proc tar_sim1(y0,b1,b2,g,e,);
local u,y,j,x,q,t,cht;
   
   t=rows(e); 
   u=e[ceil(rndu(t,1)*t)]; 
      
   y=y0|zeros(t,1); 
  j=1; 
  cht=0;
  
  for j(1,t,1);
    x=rev(y[j:_p+j-1]); 
    q=cumsumc(x[1:_delay]); q=q[_delay]; 
    x=1|x;  
    y[_p+j]=(b1'x)*(q<=g)+(b2'x)*(q>g)+u[j];
  endfor; 
  
  /*"tar(1) sim threshold=" g;
  "tar(1) sim uses betas" b1'|b2'; 
  "tar(1) sim tarres[1:10]~tarres*~TAR(1)y*" e[1:10]~(miss(0,0)|miss(0,0)|u[1:8])~y[1:10]; 
  */
   
  if maxc(abs(y)) ge 1000;
      cht=1;
  endif;
retp(cht|y); 
endp;

/*=========================================*/
/* =============== TAR2_SIM ===============*/
/*=========================================*/

proc tar2_sim(y0,b1,b2,b3,g,e);
local u,y,j,x,q,t,cht,g1,g2;
   
  /* "-------------------tar2_sim------------------";*/
   t=rows(e); 
   u=e[ceil(rndu(t,1)*t)]; 
   
   y=y0|zeros(t,1); 
   g1=g[1]; g2=g[2]; 
  
  j=1; 
  cht=0;
  do while j<=t;
    x=rev(y[j:_p+j-1]); 
     q=cumsumc(x[1:_delay]); q=q[_delay]; 
    x=1|x;  
    y[_p+j]=(b1'x)*(q<=g1)+(b2'x)*(q>g1)*(q<=g2) +(b3'x)*(q>g2)+u[j];
   j=j+1;
  endo; 
  
 /* "tar(2) sim thresholds=" g';
  "tar(2) sim uses betas" b1'|b2'|b3'; 
  "tarsim(2) tar2res[1:10]~tar2res*~AR(2)y*" e[1:10]~(miss(0,0)|miss(0,0)|u[1:8])~y[1:10]; 
 */ 
  
  
  if maxc(abs(y)) ge 1000;
      cht=1;
  endif;
retp(cht|y); 
endp;

/********* QSORT **************/
proc qsort(q);
 local qq,aarea,area;
   aarea=sortc(q,1);                    /* FIX thresh.var. */
  area=trimr(aarea,_mrs,_mrs);
  qq=unique(area,1);
 retp(qq);
endp;

proc qqsort(q);
 local qqq,aarea,area;
  aarea=sortc(q,1);                    /* FIX thresh.var. */
  if NOT rows(aarea)<2*_mrs+1;
     area=trimr(aarea,_mrs,_mrs);
     qqq=unique(area,1);
   else;
    qqq=miss(0,0);
    
  endif ;
 retp(qqq);
endp;

proc qsort_2(q,ghat);
local n,n1,qq,nq;
  n=rows(q); 
  n1=_mrs;   
  qq=unique(q,1); 
  qq=qq[n1:(rows(qq)-n1+1)]; /*qq;*/
  qq=delif(qq,(abs(seqa(1,1,rows(qq))-maxindc(qq.>=ghat)).<n1));
 retp(qq);
endp;


proc qsort_3(q,ghat);
local qq,qq1,q1,q2,q3,r1,r2,q11,q22;
  
  r1=ghat[1];
  r2=ghat[2];
  qq=unique(q,1);
  
  q1=qqsort(packr(miss(qq.*(qq.<r1),0))); /* if EXACT 0 in the dataset => gets eliminated */
  q2=qqsort(packr(miss(qq.*(qq.>r1 .and qq.<r2),0))); 
  q3=qqsort(packr(miss(qq.*(qq.>r2),0)));
    
  qq1=packr(q1|q2|q3);    
  retp(qq1);
endp;


/* ------------ Generalized inverses -------------- */
proc ginv(m);
local mi;
  trap 1;
  mi=invpd(m);
  trap 0;
  if scalerr(mi);
    mi=pinv(m);
  endif;
retp(mi);
endp;

proc dinv(y,x);
local d;
  trap 1;
    d=y/x;
  trap 0;
  if scalerr(d);
    d=ginv(moment(x,0))*(x'y);
  endif;
retp(d);
endp;


proc(1)=time_com(algus); /* need to place variable "algus" at the beginning of program */
                          /* returns time difference between algus and lopp  */
local lopp, vahe;

    lopp = time;
    vahe = lopp-algus;
    
    if vahe[4]<0;
        vahe[4] = vahe[4]+100;
        vahe[3]=vahe[3]-1;
    endif;  
    
    if vahe[3]<0;
        vahe[3] = vahe[3]+60;
        vahe[2]=vahe[2]-1;
    endif;
    
    if vahe[2]<0;
        vahe[2] = vahe[2]+60;
        vahe[1]=vahe[1]-1;
    endif;  
    
    if vahe[1]<0;
       vahe[1] = vahe[1]+24;
    endif;  
retp (timestr(vahe));
endp;

/*proc(1)=time_com2(algus); /* need to place variable "algus" at the beginning of program */
                          /* returns time difference between algus and lopp  */
local lopp, vahe,minu,tund,sec_yle,min_yle;

    lopp = timeutc;
    vahe = lopp-algus;
    
    minu=floor(vahe/60); sec_yle=vahe-minu*60;
    tund=floor(minu/60); min_yle=minu-tund*60;
    vahe=tund|min_yle|sec_yle|0;
    
retp (timestr(vahe));
endp;*/
