new;
cls;

library pgraph;
graphset;
_pdate=0;
_pnum=2;
_numht=0.7;
_pltype={6};
_plwidth=5;


chdir c:\TARSpec;

output file = c:\tarspec\citibase-LIN.OUT reset;


@ === Read data === @

load lhmu[]="c:\\TARSpec\\inpaper\\unempl\\LHMU.txt" ;      /* load raw data from LHMx.txt*/
load lhmc[]="c:\\TARSpec\\inpaper\\unempl\\LHMC.txt" ;

qt=100*(lhmu./lhmc);

qt1=trimr(qt,131,0); /* Data starting from 1959:q1 */
qt2=qt[133:rows(qt),.];
save citibase=qt2;  /* save the datamatrix NEWDATA.FMT*/

y = trimr(qt1,1,0) - trimr(lag1(qt1),1,0);     /* take 1st differences & trim */
save diffLHM=y;  /* save the datamatrix NEWDATA.FMT*/

begwind;
window(2,1,0);
setwind(1);
xy(seqa(1959,1/12,rows(qt1)),qt1);
nextwind;
xy(seqa(1959,1/12,rows(y)),y);
endwind;



p_max=18; 
jrk=seqa(0,1,p_max+1);
n_corr=12;
sig_l=0.05;
x=ones(rows(y),1);
   {AIC,p_AIC,SIC,p_SIC,LB,p_LB,pval_LB,BP,p_BP,pval_BP}=ORDCRIT(y,x,p_max,n_corr,sig_l);



meaAIC=jrk~AIC~SIC;
mask=ones(3,1);
mask=mask';
let fmt[3,3]="*.*lf " 8 4
"*.*lf " 8 4
"*.*lf " 8 4;
printfm(meaAIC,mask,fmt); 
"------------";
p_AIC~p_SIC;

ylags=lagn(y,seqa(1,1,p_AIC));
y=trimr(y,p_AIC,0);
te=rows(y);
ylag=ones(te,1)~trimr(ylags,p_AIC,0);

yi=invpd(ylag'ylag);
bhat=yi*ylag'y;
u=y-ylag*bhat;
rvar=u'u/te;
rvar1=u'u/(te-rows(bhat));
ym=meanc(y);
yxi=ones(te,1);
ssy=(y-ym*yxi)'(y-ym*yxi);
r2=1-u'u/ssy;
/* Computes model selection criteria */
yd=diag(yi);
bsd=sqrt(rvar1*yd);
tval=bhat./bsd;
sbic=ln(rvar)+(rows(bhat)*ln(te))/te;
aic=ln(rvar)+(2*rows(bhat))/te;
kn=seqa(0,1,rows(bhat));

let varstring=INTRCPT AR1 AR2 AR3 AR4 AR5 AR6 AR7 AR8 AR9 AR10 AR11 AR12;
/* Prints the results */
format /m1 8,4;
print "***********************************************************************";
print "Order of AR is    "  round(p_AIC);
print " Nr. of obs  is     "  te;
print "
variable           estimate   std.dev    t-value  " ;
meaning=varstring~bhat~bsd~tval;
mask=0|ones(3,1);
mask=mask';
let fmt[4,3]="-*.*s " 8 8
"*.*lf " 8 4
"*.*lf " 8 4
"*.*lf " 8 4;
printfm(meaning,mask,fmt); 
print " ";
print " ******** SBIC =   " sbic;
print " ********* AIC  =   " aic;
format /m0 9,7;
print "The residual sum of squares is  " u'u;
print "The residual variance is             " rvar1;
rsd=sqrt(rvar1);
print "The residual standard error is    " rsd;
rsdml=sqrt(u'u/te);
print "R^2  is                                         " r2;




/*
**  ORDCRIT.g
**
**  purpose : calculation of criteria for ARX order selection
**
**  format  :  { AIC,p_AIC,SIC,p_SIC,LB,p_LB,pval_LB,BP,p_BP,pval_BP} = ORDCRIT(y,x,p_max,n_corr,sig_l);
**
**  input   : y     : (Txk) matrix, containing time series
**            x     : (Txm) matrix, containg exogenous series (will not be
**                    lagged in proc OR 0 to indicate that no exogenous
**                    variables are present
**            p_max : scalar, maximum AR order
**            n_corr: scalar, number of residual correlations to use 
**                    in Ljung-Box test statistic
**            sig_l : scalar, significance level to use in LB test
**
**  output  : AIC   : (p_max x 1) vector, containing values of Akaike
**                    Information Criterion, cf. (4.3.2) in Lutkepohl (1991)
**            p_AIC : scalar, AR order for which AIC is minimized
**            SC    : (p_max x 1) vector, containing values of Schwarz
**                    Criterion, cf. (4.3.9) in Lutkepohl (1991)
**            p_SC  : scalar, AR order for which SC is minimized
**            LB    : (p_max x n_corr) matrix, containing values of LB test
**                    statistic. (i,j)-th element contains statistic
**                    using first j autocorrelations for residuals from
**                    ARX(i) model
**            p_LB  : (n_corr x 1), smallest AR order for which LB is 
**                    not significant at sig_l
**            BP    : Breusch-Pagan test of autocorrelation 
**                    its p-value
**
**  globals :
**
**  remarks :
**
**  authors : Dick van Dijk and Stefan Lundbergh
**
**  date    : First attempt on 13 October 1999
**            Last revision on 13 October 1999
**
*/
proc (10) = ORDCRIT(y,x,p_max,n_corr,sig_l);
  local AIC,SIC,LB,pval_LB,BP,pval_BP,T,k_x,_dat,_y,_x,i,phi_l,_y_fit,_e_fit,tx,corr,Tk,
        p_AIC,p_SIC,p_LB,p_BP,j,t0,t1,t2,t3;

AIC=zeros(p_max+1,1);
SIC=zeros(p_max+1,1);
LB=zeros(p_max+1,n_corr);
pval_LB=zeros(p_max+1,n_corr);
BP=zeros(p_max+1,n_corr);
pval_BP=zeros(p_max+1,n_corr);

T=rows(y)-p_max;
k_x=cols(x);
_dat=y[p_max+1:T+p_max]~x[p_max+1:T+p_max,.]~
     trimr(lagn(y,seqa(1,1,p_max)),p_max,0);
_y=_dat[.,1];

i=-1;
do until (i==p_max);
  i=i+1;
  _x=_dat[.,2:2+k_x+i-1];

  /* Estimate ARX(p) model */
  phi_l=invpd(_x'_x)*(_x'_y);

  /* Compute output, residuals, AIC, SIC, residual correlations and LB statistic*/
  _y_fit=_x*phi_l;
  _e_fit=_y-_y_fit;
  AIC[i+1]=ln(meanc(_e_fit.^2))+2*(i+k_x)./T;
  SIC[i+1]=ln(meanc(_e_fit.^2))+(i+k_x)*ln(T)./T;

  tx=missrv(lagn(_e_fit,seqa(0,1,n_corr+1)),0);
  corr=((tx[.,1]'tx[.,2:n_corr+1])./(tx[.,1]'tx[.,1]))';

  Tk=seqa(T-1,-1,n_corr);
  LB[i+1,.]=T*(T+2)*cumsumc((corr.^2)./Tk)';
  /* NOTE: degrees of freedom is corrected for number of estimated parameters in ARX model */
  if (i+k_x .LE n_corr);
    pval_LB[i+1,i+k_x:n_corr]=CDFCHIC(LB[i+1,i+k_x:n_corr],seqa(1,1,n_corr-i-k_x+1)');
  endif;

  j=0;
  do until (j==n_corr);
    j=j+1;
    BP[i+1,j]=T*(tx[.,1]'(_x~tx[.,2:j+1]))*
                    invpd((_x~tx[.,2:j+1])'(_x~tx[.,2:j+1]))*
                   ((_x~tx[.,2:j+1])'tx[.,1])./(tx[.,1]'tx[.,1]); 

    pval_BP[i+1,j]=CDFCHIC(BP[i+1,j],j);
  endo;

endo;

p_AIC=minindc(AIC)-1;
p_SIC=minindc(SIC)-1;
p_LB=zeros(n_corr,1);
p_BP=zeros(n_corr,1);
i=0;
do until (i==n_corr);
  i=i+1; 
  if (sumc(pval_LB[.,i] .GT sig_l)==0); 
    p_LB[i]=p_max;
  else;
    p_LB[i]=maxindc(pval_LB[.,i] .GT sig_l) - 1;
  endif;
  if (sumc(pval_BP[.,i] .GT sig_l)==0); 
    p_BP[i]=p_max;
  else;
    p_BP[i]=maxindc(pval_BP[.,i] .GT sig_l) - 1;
  endif;
endo;

retp(AIC,p_AIC,SIC,p_SIC,LB,p_LB,pval_LB,BP,p_BP,pval_BP);
endp;
